/*
 *  Copyright (C) 2011 Christian Werner <chw@ch-werner.de>.
 *  All Rights Reserved.
 *
 *  This is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 */

/*
 * tkVNC.c --
 *
 *	Tk VNC widget
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <fcntl.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/XShm.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <tk.h>
#if (TK_MAJOR_VERSION < 8)
#error "need Tk 8.0 or higher"
#endif
#if (TK_MAJOR_VERSION == 8) && (TK_MINOR_VERSION == 0)
#define TK80 1
#endif
#ifdef HAVE_PTH
#include <pth.h>
#endif

#define KEYSYM_H

#include "rfb.h"
#include "rfbclient.h"

#define SHM_WIDTH  1280
#define SHM_HEIGHT 1024
#define SHM_SIZE (SHM_WIDTH * SHM_HEIGHT * 4)

/*
 * Flag bits for VNC:
 *
 * REDRAW_PENDING:		Non-zero means a DoWhenIdle handler
 *				has already been queued to redraw
 *				this window.
 * COMPL_PENDING:
 */

#define REDRAW_PENDING		1
#define REDRAW_XEVENTS		2
#define REDRAW_NOCLIP		2
#define COMPL_PENDING		8
#define REDRAW_AGAIN		16
#define GOT_FOCUS		32
#define REQUEST_REDRAW		64
#define UPDATE_XVIEW		128
#define UPDATE_YVIEW		256
#define UPDATE_INFO		512

/*
 * A data structure of the following type is kept for each
 * VNC that currently exists for this process:
 */

typedef struct {
    Tk_Window tkwin;		/* Window that embodies the VNC.  NULL
				 * means that the window has been destroyed
				 * but the data structures haven't yet been
				 * cleaned up. */
    Display *display;		/* Display containing widget.  Used, among
				 * other things, so that resources can be
				 * freed even after tkwin has gone away. */
    Tcl_Interp *interp;		/* Interpreter associated with widget. Used
				 * to delete widget command. */
    Tcl_Command widgetCmd;	/* Token for VNC's widget command. */
    Tk_Cursor cursor;		/* Current cursor for window, or None. */
    char *takeFocus;		/* Value of -takefocus option;  not used in
				 * the C code, but used by keyboard traversal
				 * scripts.  Malloc'ed, but may be NULL. */
    Tk_3DBorder border;		/* Structure used to draw 3-D border.
				 * NULL means no border. */
    int borderWidth;		/* Width of 3-D border (if any). */
    int relief;			/* 3-D effect: TK_RELIEF_RAISED etc. */
    int width;			/* Width to request for window.  <= 0 means
				 * don't request any size. */
    int height;			/* Height to request for window.  <= 0 means
				 * don't request any size. */
    int highlightWidth;		/* Width in pixels of highlight to draw
				 * around widget when it has the focus.
				 * 0 means don't draw a highlight. */
    XColor *highlightBgColorPtr;/* Color for drawing traversal highlight
				 * area when highlight is off. */
    XColor *highlightColorPtr;	/* Color for drawing traversal highlight. */
    XColor *fillColorPtr;	/* Color for background filling. */
    char *xScrollCmd;		/* Command prefix for communicating with
				 * horizontal scrollbar.  NULL means no command
				 * to issue.  Malloc'ed. */
    char *yScrollCmd;		/* Command prefix for communicating with
				 * vertical scrollbar.  NULL means no command
				 * to issue.  Malloc'ed. */
    char *infoCmd;		/* Command prefix for connection state
				 * callback procedure.  NULL means no command
				 * to issue.  Malloc'ed. */

    GC iGC;			/* GC for drawing XImage. */
    Region region;		/* Region to optimize expose/XPutImage. */
    int islocal;		/* True for local X display connection. */
    int canlocal;		/* True for using XShmPutImage. */
    XShmSegmentInfo shmInfo;	/* XShm IPC descriptor. */
    int Completion;		/* XShm completion event number. */
    XImage *img[2];		/* XImage for VNC frame buffer. */
    int xoff, yoff;		/* X/Y view offsets for scrolling. */
    int vxoff, vyoff;		/* Computed view offsets for drawing. */
    int viw, vih;		/* Ditto, image dimension. */
    int needPwd;		/* True when password required. */
    char *pwd;			/* VNC password, malloc'ed. */
    int flags;			/* Various flags;  see above for
				 * definitions. */
#ifdef HAVE_PTH
    pth_uctx_t uc0, uc1;	/* Contexts for coroutine handling. */
    int uc_run;			/* True when uc1 is running. */
#endif

    rfbClient rfb;		/* RFB client structure from rfbclient.h. */ 
} VNC;

static Tk_ConfigSpec configSpecs[] = {
    {TK_CONFIG_BORDER, "-background", "background", "Background",
	"#d9d9d9", Tk_Offset(VNC, border), 0},
    {TK_CONFIG_SYNONYM, "-bd", "borderWidth", (char *) NULL,
	(char *) NULL, 0, 0},
    {TK_CONFIG_PIXELS, "-borderwidth", "borderWidth", "BorderWidth",
	"0", Tk_Offset(VNC, borderWidth), 0},
    {TK_CONFIG_SYNONYM, "-bg", "background", (char *) NULL,
       (char *) NULL, 0, 0},
    {TK_CONFIG_ACTIVE_CURSOR, "-cursor", "cursor", "Cursor",
	(char *) NULL, Tk_Offset(VNC, cursor), TK_CONFIG_NULL_OK},
    {TK_CONFIG_COLOR, "-fill", "fillColor", "FillColor",
	(char *) NULL, Tk_Offset(VNC, fillColorPtr), TK_CONFIG_NULL_OK},
    {TK_CONFIG_PIXELS, "-height", "height", "Height",
	"0", Tk_Offset(VNC, height), 0},
    {TK_CONFIG_COLOR, "-highlightbackground", "highlightBackground",
	"HighlightBackground", "#d9d9d9", Tk_Offset(VNC, highlightBgColorPtr),
	0},
    {TK_CONFIG_COLOR, "-highlightcolor", "highlightColor", "HighlightColor",
	"#000000", Tk_Offset(VNC, highlightColorPtr), 0},
    {TK_CONFIG_PIXELS, "-highlightthickness", "highlightThickness",
	"HighlightThickness", "0", Tk_Offset(VNC, highlightWidth), 0},
    {TK_CONFIG_STRING, "-infocommand", "infoCommand", "InfoCommand",
	NULL, Tk_Offset(VNC, infoCmd), TK_CONFIG_NULL_OK},
    {TK_CONFIG_RELIEF, "-relief", "relief", "Relief",
	"flat", Tk_Offset(VNC, relief), 0},
    {TK_CONFIG_STRING, "-takefocus", "takeFocus", "TakeFocus",
	"1", Tk_Offset(VNC, takeFocus), TK_CONFIG_NULL_OK},
    {TK_CONFIG_PIXELS, "-width", "width", "Width",
	"0", Tk_Offset(VNC, width), 0},
    {TK_CONFIG_STRING, "-xscrollcommand", "xScrollCommand", "ScrollCommand",
	NULL, Tk_Offset(VNC, xScrollCmd), TK_CONFIG_NULL_OK},
    {TK_CONFIG_STRING, "-yscrollcommand", "yScrollCommand", "ScrollCommand",
	NULL, Tk_Offset(VNC, yScrollCmd), TK_CONFIG_NULL_OK},
    {TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
	(char *) NULL, 0, 0}
};

/*
 * Forward declarations for procedures defined later in this file:
 */

static void		RFBBell _ANSI_ARGS_((rfbClient *rfbPtr));
#ifdef HAVE_PTH
static rfbBool		RFBSwitch _ANSI_ARGS_((rfbClient *rfbPtr));
static void		DoRFBServerMessage _ANSI_ARGS_((void *arg));
#endif
static rfbBool		RFBMallocFrameBuffer _ANSI_ARGS_((rfbClient *rfbPtr));
static rfbBool		RFBCursorPos _ANSI_ARGS_((rfbClient *rfbPtr,
						  int x, int y));
static void		RFBCursorLockArea _ANSI_ARGS_((rfbClient *rfbPtr,
						       int x, int y,
						       int w, int h));
static void		RFBCursorUnlockScreen _ANSI_ARGS_((rfbClient *rfbPtr));
static void		RFBFrameBufferUpdate _ANSI_ARGS_((rfbClient *rfbPtr,
							  int x, int y,
							  int w, int h));
static char *		RFBReadPassword _ANSI_ARGS_((rfbClient *rfbPtr));
static void		RFBInitClient _ANSI_ARGS_((rfbClient *rfbPtr,
						   ClientData clientData,
						   int bytesPerPixel,
						   int bitsPerSample,
						   int samplesPerPixel,
						   Visual *vi));
static rfbBool		RFBInitConnection _ANSI_ARGS_((rfbClient *rfbPtr,
						       int shared,
						       int viewOnly));
static void		RFBCloseConnection _ANSI_ARGS_((rfbClient *rfbPtr));
static void		VNCFileHandler _ANSI_ARGS_((ClientData clientData,
						    int mask));

static int		HandleCompletionEvents _ANSI_ARGS_((ClientData clientData,
							    XEvent *eventPtr));
static int		ShmErrHandler _ANSI_ARGS_((ClientData clientData,
						   XErrorEvent *errEvent));
static int		VNCCmd _ANSI_ARGS_((ClientData clientData,
					    Tcl_Interp *interp, int argc,
					    char **argv));
static void		MakeInfo _ANSI_ARGS_((VNC *vncPtr,
					      Tcl_DString *dsPtr));
static int		VNCWidgetCmd _ANSI_ARGS_((ClientData clientData,
						  Tcl_Interp *interp,
						  int argc, char **argv));
static void		DestroyVNC _ANSI_ARGS_((char *memPtr));
static int		ConfigureVNC _ANSI_ARGS_((Tcl_Interp *interp,
						  VNC *vncPtr, int argc,
						  char **argv, int flags));
static void		ComputeView _ANSI_ARGS_((VNC *vncPtr));
static void		DisplayVNC _ANSI_ARGS_((ClientData clientData));
static void		VNCEventProc _ANSI_ARGS_((ClientData clientData,
						  XEvent *eventPtr));
static void		VNCCmdDeletedProc _ANSI_ARGS_((ClientData clientData));



/*
 *--------------------------------------------------------------
 *
 * RFBBell --
 *
 *	Ring X bell.
 *
 *--------------------------------------------------------------
 */

static void
RFBBell(rfbPtr)
    rfbClient *rfbPtr;
{
    VNC *vncPtr = (VNC *) rfbPtr->clientData;

    if (vncPtr->display) {
	XBell(vncPtr->display, 100);
    }
}
#ifdef HAVE_PTH

/*
 *--------------------------------------------------------------
 *
 * RFBSwitch --
 * DoRFBServerMessage --
 *
 *
 *--------------------------------------------------------------
 */

static rfbBool
RFBSwitch(rfbPtr)
    rfbClient *rfbPtr;
{
    VNC *vncPtr = (VNC *) rfbPtr->clientData;

    if (!vncPtr->uc_run) {
	return FALSE;
    }
    pth_uctx_switch(vncPtr->uc1, vncPtr->uc0);
    return TRUE;
}

static void
DoRFBServerMessage(arg)
    void *arg;
{
    VNC *vncPtr = (VNC *) arg;

    while (1) {
	while (RFBHandleRFBServerMessage(&vncPtr->rfb)) {
	    /* nothing */
	}
	RFBCloseConnection(&vncPtr->rfb);
	vncPtr->uc_run = 0;
	pth_uctx_switch(vncPtr->uc1, vncPtr->uc0);
    }
}
#endif

/*
 *--------------------------------------------------------------
 *
 * RFBMallocFrameBuffer --
 *
 *
 *--------------------------------------------------------------
 */

static rfbBool
RFBMallocFrameBuffer(rfbPtr)
    rfbClient *rfbPtr;
{
    VNC *vncPtr = (VNC *) rfbPtr->clientData;
    int size, i;

    size = rfbPtr->width * rfbPtr->height * rfbPtr->format.bitsPerPixel / 8;
    if (vncPtr->islocal) {
	if (size < SHM_SIZE) {
	    if (!vncPtr->canlocal && rfbPtr->frameBuffer) {
		Tcl_Free((char *) rfbPtr->frameBuffer);
	    }
	    rfbPtr->frameBuffer = (unsigned char *) vncPtr->shmInfo.shmaddr;
	    vncPtr->canlocal = i = 1;
	} else {
	    goto nolocal;
	}
    } else {
nolocal:
	if (!vncPtr->canlocal && rfbPtr->frameBuffer) {
	    Tcl_Free((char *) rfbPtr->frameBuffer);
	}
	rfbPtr->frameBuffer = (unsigned char *) Tcl_Alloc(size);
	vncPtr->canlocal = i = 0;
    }
    vncPtr->img[i]->bytes_per_line = rfbPtr->width *
	rfbPtr->format.bitsPerPixel / 8;
    vncPtr->img[i]->width = rfbPtr->width;
    vncPtr->img[i]->height = rfbPtr->height;
    vncPtr->img[i]->data = (char *) rfbPtr->frameBuffer;
    if (vncPtr->width <= 0 || vncPtr->height <= 0) {
	int width, height;

	width = rfbPtr->width + 
	    2 * (vncPtr->highlightWidth + vncPtr->borderWidth);
	height = rfbPtr->height + 
	    2 * (vncPtr->highlightWidth + vncPtr->borderWidth);
	Tk_GeometryRequest(vncPtr->tkwin, width, height);
	ComputeView(vncPtr);
	vncPtr->flags |= UPDATE_XVIEW | UPDATE_YVIEW | UPDATE_INFO;
    }
    if (rfbPtr->frameBuffer) {
	return TRUE;
    }
    return FALSE;
}

/*
 *--------------------------------------------------------------
 *
 * RFBCursorPos --
 *
 *
 *--------------------------------------------------------------
 */

static rfbBool
RFBCursorPos(rfbPtr, x, y)
    rfbClient *rfbPtr;
    int x, y;
{
    return TRUE;
}

/*
 *--------------------------------------------------------------
 *
 * RFBCursorLockArea --
 *
 *
 *--------------------------------------------------------------
 */

static void
RFBCursorLockArea(rfbPtr, x, y, w, h)
    rfbClient *rfbPtr;
    int x, y, w, h;
{
}

/*
 *--------------------------------------------------------------
 *
 * RFBCursorUnlockScreen --
 *
 *
 *--------------------------------------------------------------
 */

static void
RFBCursorUnlockScreen(rfbPtr)
    rfbClient *rfbPtr;
{
}

/*
 *--------------------------------------------------------------
 *
 * RFBFrameBufferUpdate --
 *
 *
 *--------------------------------------------------------------
 */

static void
RFBFrameBufferUpdate(rfbPtr, x, y, w, h)
    rfbClient *rfbPtr;
    int x, y, w, h;
{
    VNC *vncPtr = (VNC *) rfbPtr->clientData;
    XRectangle rect;

    rect.x = x + vncPtr->vxoff - vncPtr->xoff;
    rect.y = y + vncPtr->vyoff - vncPtr->yoff;
    rect.width = w;
    rect.height = h;
    XUnionRectWithRegion(&rect, vncPtr->region, vncPtr->region);
#ifdef HAVE_PTH
    vncPtr->flags |= REQUEST_REDRAW;
#else
    if (!(vncPtr->flags & REDRAW_PENDING)) {
	Tcl_DoWhenIdle(DisplayVNC, (ClientData) vncPtr);
	vncPtr->flags |= REDRAW_PENDING;
    }
#endif
}

/*
 *--------------------------------------------------------------
 *
 * RFBReadPassword --
 *
 *
 *--------------------------------------------------------------
 */

static char *
RFBReadPassword(rfbPtr)
    rfbClient *rfbPtr;
{
    VNC *vncPtr = (VNC *) rfbPtr->clientData;

    if (vncPtr->pwd == NULL) {
	vncPtr->needPwd = 1;
	return NULL;
    }
    return strdup(vncPtr->pwd);
}

/*
 *--------------------------------------------------------------
 *
 * RFBInitClient --
 *
 *
 *--------------------------------------------------------------
 */

static void
RFBInitClient(rfbPtr, clientData,
	      bytesPerPixel, bitsPerSample, samplesPerPixel,
	      vi)
    rfbClient *rfbPtr;
    ClientData clientData;
    int bytesPerPixel;
    int bitsPerSample;
    int samplesPerPixel;
    Visual *vi;
{
    AppData *data = &rfbPtr->appData;

    data->shareDesktop = FALSE;
    data->viewOnly = FALSE;
#if defined(HAVE_LIBJPEG) && defined(HAVE_LIBZ)
    data->encodingsString = "tight hextile zlib corre rre raw";
#elif defined(HAVE_LIBZ)
    data->encodingsString = "hextile zlib corre rre raw";
#elif defined(HAVE_LIBJPEG)
    data->encodingsString = "tight hextile corre rre raw";
#else
    data->encodingsString = "hextile corre rre raw";
#endif
    data->useBGR233 = FALSE;
    data->nColours = 0;
    data->forceOwnCmap = FALSE;
    data->forceTrueColour = FALSE;
    data->requestedDepth = 0;
    data->compressLevel = 3;
    data->qualityLevel = 5;
#ifdef HAVE_LIBJPEG
    data->enableJPEG = TRUE;
#else
    data->enableJPEG = FALSE;
#endif
    data->useRemoteCursor = FALSE;

    rfbPtr->sock = -1;
    rfbPtr->endianTest = 1;
    rfbPtr->programName= "";
    rfbPtr->serverHost = NULL;
    rfbPtr->serverPort = 5900;

    rfbPtr->format.bitsPerPixel = bytesPerPixel * 8;
    rfbPtr->format.depth = bitsPerSample * samplesPerPixel;
    rfbPtr->appData.requestedDepth = rfbPtr->format.depth;
    rfbPtr->format.bigEndian =
	((char *) &rfbPtr->endianTest)[0] ? FALSE : TRUE;
    rfbPtr->format.trueColour = TRUE;

    if (rfbPtr->format.bitsPerPixel == 8) {
	rfbPtr->format.redMax = 7;
	rfbPtr->format.greenMax = 7;
	rfbPtr->format.blueMax = 3;
	rfbPtr->format.redShift = 0;
	rfbPtr->format.greenShift = 3;
	rfbPtr->format.blueShift = 6;
    } else {
	rfbPtr->format.redMax = (1 << bitsPerSample) - 1;
	rfbPtr->format.greenMax = (1 << bitsPerSample) - 1;
	rfbPtr->format.blueMax = (1 << bitsPerSample) - 1;
	if (vi) {
	    if (bitsPerSample >= 8) {
		if (vi->red_mask & 0x80000000) {
		    rfbPtr->format.redShift = 24;
		} else if (vi->red_mask & 0x00800000) {
		    rfbPtr->format.redShift = 16;
		} else if (vi->red_mask & 0x00008000) {
		    rfbPtr->format.redShift = 8;
		} else {
		    rfbPtr->format.redShift = 0;
		}
		if (vi->green_mask & 0x80000000) {
		    rfbPtr->format.greenShift = 24;
		} else if (vi->green_mask & 0x00800000) {
		    rfbPtr->format.greenShift = 16;
		} else if (vi->green_mask & 0x00008000) {
		    rfbPtr->format.greenShift = 8;
		} else {
		   rfbPtr->format.greenShift = 0;
		}
		if (vi->blue_mask & 0x80000000) {
		    rfbPtr->format.blueShift = 24;
		} else if (vi->blue_mask & 0x00800000) {
		    rfbPtr->format.blueShift = 16;
		} else if (vi->blue_mask & 0x00008000) {
		    rfbPtr->format.blueShift = 8;
		} else {
		    rfbPtr->format.blueShift = 0;
		}
	    } else {
		if (vi->red_mask & 0x8000) {
		    rfbPtr->format.redShift = 11;
		} else if (vi->red_mask & 0x4000) {
		    rfbPtr->format.redShift = 10;
		} else if (vi->red_mask & 0x0400) {
		    rfbPtr->format.redShift = 6;
		} else if (vi->red_mask & 0x0200) {
		    rfbPtr->format.redShift = 5;
		} else {
		    rfbPtr->format.redShift = 0;
		}
		if (vi->green_mask & 0x8000) {
		    rfbPtr->format.greenShift = 11;
		} else if (vi->green_mask & 0x4000) {
		    rfbPtr->format.greenShift = 10;
		} else if (vi->green_mask & 0x0400) {
		    rfbPtr->format.greenShift = 6;
		} else if (vi->green_mask & 0x0200) {
		    rfbPtr->format.greenShift = 5;
		} else {
		    rfbPtr->format.greenShift = 0;
		}
		if (vi->blue_mask & 0x8000) {
		    rfbPtr->format.blueShift = 11;
		} else if (vi->blue_mask & 0x4000) {
		    rfbPtr->format.blueShift = 10;
		} else if (vi->blue_mask & 0x0400) {
		    rfbPtr->format.blueShift = 6;
		} else if (vi->blue_mask & 0x0200) {
		    rfbPtr->format.blueShift = 5;
		} else {
		    rfbPtr->format.blueShift = 0;
		}
	    }
	} else if (!rfbPtr->format.bigEndian) {
	    rfbPtr->format.redShift = 0;
	    rfbPtr->format.greenShift = bitsPerSample;
	    rfbPtr->format.blueShift = bitsPerSample * 2;
	} else {
	    if (rfbPtr->format.bitsPerPixel == 8 * 3) {
		rfbPtr->format.redShift = bitsPerSample * 2;
		rfbPtr->format.greenShift = bitsPerSample * 1;
		rfbPtr->format.blueShift = 0;
	    } else {
		rfbPtr->format.redShift = bitsPerSample * 3;
		rfbPtr->format.greenShift = bitsPerSample * 2;
		rfbPtr->format.blueShift = bitsPerSample;
	    }
	}
    }

    rfbPtr->bufoutptr = rfbPtr->buf;
    rfbPtr->buffered = 0;

#ifdef HAVE_LIBZ
    rfbPtr->raw_buffer_size = -1;
    rfbPtr->raw_buffer = NULL;
    rfbPtr->decompStreamInited = FALSE;
#endif

#ifdef HAVE_LIBJPEG
    memset(rfbPtr->zlibStreamActive, 0, sizeof (rfbBool) * 4);
    rfbPtr->jpegSrcManager = NULL;
#endif

    rfbPtr->HandleCursorPos = RFBCursorPos;
    rfbPtr->SoftCursorLockArea = RFBCursorLockArea;
    rfbPtr->SoftCursorUnlockScreen = RFBCursorUnlockScreen;
    rfbPtr->GotFrameBufferUpdate = RFBFrameBufferUpdate;
    rfbPtr->GetPassword = RFBReadPassword;
    rfbPtr->MallocFrameBuffer = RFBMallocFrameBuffer;
    rfbPtr->Bell = RFBBell;
    rfbPtr->Suspend = NULL;

    rfbPtr->clientData = (void *) clientData;
}

/*
 *--------------------------------------------------------------
 *
 * RFBInitConnection --
 *
 *
 *--------------------------------------------------------------
 */

static rfbBool
RFBInitConnection(rfbPtr, shared, viewOnly)
    rfbClient *rfbPtr;
    int shared;
    int viewOnly;
{
    int val;
#ifndef _WIN32
    struct timeval tv;
#endif
    AppData *data = &rfbPtr->appData;

    data->shareDesktop = shared ? TRUE : FALSE;
    data->viewOnly = viewOnly ? TRUE : FALSE;

    rfbPtr->bufoutptr = rfbPtr->buf;
    rfbPtr->buffered = 0;

    if (!rfbPtr->serverHost ||
	!RFBConnectToRFBServer(rfbPtr, 
			    rfbPtr->serverHost, rfbPtr->serverPort)) {
	return FALSE;
    }
    val = 1;
    setsockopt(rfbPtr->sock, SOL_SOCKET, SO_KEEPALIVE, (char *) &val,
	       sizeof (val));
#ifdef TCP_KEEPIDLE
    val = 30;
    setsockopt(rfbPtr->sock, SOL_TCP, TCP_KEEPIDLE, (char *) &val,
	       sizeof (val));
#endif
#ifdef TCP_KEEPINTVL
    val = 30;
    setsockopt(rfbPtr->sock, SOL_TCP, TCP_KEEPINTVL, (char *) &val,
	       sizeof (val));
#endif
#ifdef TCP_KEEPCNT
    val = 10;
    setsockopt(rfbPtr->sock, SOL_TCP, TCP_KEEPCNT, (char *) &val,
	       sizeof (val));
#endif
#ifdef _WIN32
    val = 20000;
    setsockopt(rfbPtr->sock, SOL_SOCKET, SO_RCVTIMEO, (char *) &val,
	       sizeof (val));
    val = 20000;
    setsockopt(rfbPtr->sock, SOL_SOCKET, SO_SNDTIMEO, (char *) &val,
	       sizeof (val));
#else
    tv.tv_sec = 20;
    tv.tv_usec = 0;
    setsockopt(rfbPtr->sock, SOL_SOCKET, SO_RCVTIMEO, (char *) &tv,
	       sizeof (tv));
    tv.tv_sec = 20;
    tv.tv_usec = 0;
    setsockopt(rfbPtr->sock, SOL_SOCKET, SO_SNDTIMEO, (char *) &tv,
	       sizeof (tv));
#endif
    if (!RFBInitialiseRFBConnection(rfbPtr)) {
	return FALSE;
    }
    if (!RFBSetFormatAndEncodings(rfbPtr)) {
	return FALSE;
    }
    rfbPtr->width = rfbPtr->si.framebufferWidth;
    rfbPtr->height = rfbPtr->si.framebufferHeight;
    rfbPtr->MallocFrameBuffer(rfbPtr);
    if (!RFBSendFramebufferUpdateRequest(rfbPtr, 0, 0,
					 rfbPtr->width, rfbPtr->height,
					 FALSE)) {
	return FALSE;
    }
    return TRUE;
}

/*
 *--------------------------------------------------------------
 *
 * RFBCloseConnection --
 *
 *
 *--------------------------------------------------------------
 */

static void
RFBCloseConnection(rfbPtr)
    rfbClient *rfbPtr;
{
    if (rfbPtr->sock != -1) {
	close(rfbPtr->sock);
	rfbPtr->sock = -1;
    }
}

/*
 *--------------------------------------------------------------
 *
 * VNCFileHandler --
 *
 *
 *--------------------------------------------------------------
 */

static void
VNCFileHandler(clientData, mask)
    ClientData clientData;
    int mask;
{
    VNC *vncPtr = (VNC *) clientData;

    if ((mask & TCL_READABLE) && vncPtr->rfb.sock != -1) {
#ifdef HAVE_PTH
	int redraw = 0;

	Tcl_DeleteFileHandler(vncPtr->rfb.sock);
	pth_uctx_switch(vncPtr->uc0, vncPtr->uc1);
	if (vncPtr->rfb.sock == -1) {
	    redraw = 1;
	    vncPtr->flags |= UPDATE_XVIEW | UPDATE_YVIEW | UPDATE_INFO;
	} else {
	    redraw = (vncPtr->flags & REQUEST_REDRAW) != 0;
	    Tcl_CreateFileHandler(vncPtr->rfb.sock, TCL_READABLE,
				  VNCFileHandler, (ClientData) vncPtr);
	}
	vncPtr->flags &= ~REQUEST_REDRAW;
	if (redraw && !(vncPtr->flags & REDRAW_PENDING)) {
	    Tcl_DoWhenIdle(DisplayVNC, (ClientData) vncPtr);
	    vncPtr->flags |= REDRAW_PENDING;
	}
#else
	if (!RFBHandleRFBServerMessage(&vncPtr->rfb)) {
	    Tcl_DeleteFileHandler(vncPtr->rfb.sock);
	    RFBCloseConnection(&vncPtr->rfb);
	    vncPtr->flags |= UPDATE_XVIEW | UPDATE_YVIEW | UPDATE_INFO;
	    if (!(vncPtr->flags & REDRAW_PENDING)) {
		Tcl_DoWhenIdle(DisplayVNC, (ClientData) vncPtr);
		vncPtr->flags |= REDRAW_PENDING;
	    }
	}
#endif
    }
}

/*
 *--------------------------------------------------------------
 *
 * HandleCompletionEvents --
 *
 *
 *--------------------------------------------------------------
 */

static int
HandleCompletionEvents(clientData, eventPtr)
    ClientData clientData;
    XEvent *eventPtr;
{
    VNC *vncPtr = (VNC *) clientData;
    Tk_Window tkwin = vncPtr->tkwin;
    XShmCompletionEvent *shmEvent = (XShmCompletionEvent *) eventPtr;

    if (eventPtr->xany.type == vncPtr->Completion &&
	vncPtr->shmInfo.shmseg == shmEvent->shmseg) {
	if (tkwin == NULL || !Tk_IsMapped(tkwin)) {
	    vncPtr->flags &= ~(COMPL_PENDING | REDRAW_AGAIN);
	    return 1;
	}
	vncPtr->flags &= ~COMPL_PENDING;
	if (vncPtr->flags & REDRAW_AGAIN) {
	    if (!(vncPtr->flags & REDRAW_PENDING)) {
		Tcl_DoWhenIdle(DisplayVNC, (ClientData) vncPtr);
		vncPtr->flags |= REDRAW_PENDING;
	    }
	    vncPtr->flags &= ~REDRAW_AGAIN;
	}
	return 1;
    }
    return 0;
}

/*
 *--------------------------------------------------------------
 *
 * ShmErrHandler --
 *
 *	X error handler during XShm image creation.
 *
 *--------------------------------------------------------------
 */

static int
ShmErrHandler(clientData, errEvent)
    ClientData clientData;
    XErrorEvent *errEvent;
{
    int *goterr = (int *) clientData;

    *goterr = 1;
    return 0;
}

/*
 *--------------------------------------------------------------
 *
 * VNCCmd --
 *
 *	This procedure is invoked to process the "vnc"
 *	Tcl command.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

static int
VNCCmd(clientData, interp, argc, argv)
    ClientData clientData;	/* Main window associated with interpreter.
				 * If we're called by Tk_Init to create a
				 * new application, then this is NULL. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int argc;			/* Number of arguments. */
    char **argv;		/* Argument strings. */
{
    Tk_Window tkwin = (Tk_Window) clientData;
    VNC *vncPtr;
    Tk_Window new;
    int bypp, bps, spp, islocal = 0;
    XGCValues gcValues;
    Visual *vi;
    int mask, screen;
    socklen_t paddrLen, laddrLen;
    struct sockaddr paddr, laddr;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
		argv[0], " pathName ?options?\"", (char *) NULL);
	return TCL_ERROR;
    }

    switch (Tk_Depth(tkwin)) {
    case 15:
    case 16:
	bypp = 2;
	bps = 5;
	spp = 3;
	break;
    case 24:
    case 32:
	bypp = 4;
	bps = 8;
	spp = 3;
	break;
    default:
	Tcl_AppendResult(interp, "unsupported display depth", (char *) NULL);
	return TCL_ERROR;
    }

    /* Find out if we are on a local display. */
    paddrLen = sizeof (paddr);
    memset(&paddr, 0, sizeof (paddr));
    laddrLen = sizeof (laddr);
    memset(&laddr, 0, sizeof (laddr));
    if (getpeername(ConnectionNumber(Tk_Display(tkwin)),
				     &paddr, &paddrLen) < 0 ||
	getsockname(ConnectionNumber(Tk_Display(tkwin)),
				     &laddr, &laddrLen) < 0) {
	Tcl_AppendResult(interp, "cannot determine X server connection: ",
			 Tcl_PosixError(interp), (char *) NULL);
	return TCL_ERROR;
    }
    switch (paddr.sa_family) {
    case AF_INET:
	if (((struct sockaddr_in *) &paddr)->sin_addr.s_addr ==
	    ((struct sockaddr_in *) &laddr)->sin_addr.s_addr) {
	    if (ntohs(((struct sockaddr_in *) &paddr)->sin_port) < 6010) {
		/* assume SSH starts at localhost:10 */
		islocal = 1;
	    }
	}
	break;
    case AF_UNIX:
	islocal = 1;
	break;
    }

    Tk_MakeWindowExist(tkwin);
    new = Tk_CreateWindowFromPath(interp, tkwin, argv[1], (char *) NULL);
    if (new == NULL) {
	goto error;
    }
    Tk_SetClass(new, "VNC");

    vncPtr = (VNC *) Tcl_Alloc(sizeof (VNC));
    memset(vncPtr, 0, sizeof (VNC));
    vncPtr->islocal = islocal;
    vncPtr->canlocal = 0;
    vncPtr->tkwin = new;
    vncPtr->display = Tk_Display(new);
    vncPtr->interp = interp;
    vncPtr->widgetCmd = Tcl_CreateCommand(interp,
					  Tk_PathName(new), VNCWidgetCmd,
					  (ClientData) vncPtr,
					  VNCCmdDeletedProc);
    vncPtr->cursor = None;
    vncPtr->takeFocus = NULL;
    vncPtr->border = NULL;
    vncPtr->borderWidth = 0;
    vncPtr->relief = TK_RELIEF_FLAT;
    vncPtr->width = 0;
    vncPtr->height = 0;
    vncPtr->highlightWidth = 0;
    vncPtr->highlightBgColorPtr = NULL;
    vncPtr->highlightColorPtr = NULL;
    vncPtr->fillColorPtr = NULL;
    vncPtr->iGC = None;
    vncPtr->flags = 0;
    vncPtr->xoff = vncPtr->yoff = 0;
    vncPtr->vxoff = vncPtr->vyoff = 0;
    vncPtr->vih = vncPtr->vih = 0;

    screen = Tk_ScreenNumber(tkwin);
    vncPtr->region = XCreateRegion();

    Tk_MakeWindowExist(new);

    gcValues.foreground = BlackPixel(vncPtr->display, screen);
    gcValues.clip_x_origin = 0;
    gcValues.clip_y_origin = 0;
    gcValues.graphics_exposures = True;
    mask = GCForeground | GCClipXOrigin | GCClipYOrigin | GCGraphicsExposures;
    vncPtr->iGC = XCreateGC(vncPtr->display, Tk_WindowId(vncPtr->tkwin),
			    mask, &gcValues);

    vi = DefaultVisual(vncPtr->display, screen);
    if (islocal && XShmQueryExtension(vncPtr->display)) {
	Tk_ErrorHandler handler;
	int goterr = 0;

	handler = Tk_CreateErrorHandler(vncPtr->display, -1, -1, -1,
					ShmErrHandler, (ClientData) &goterr);
	vncPtr->Completion = XShmGetEventBase(vncPtr->display) + ShmCompletion;
	if (goterr) {
	    Tk_DeleteErrorHandler(handler);
	    goto noshm_2;
	}
	vncPtr->img[1] = XShmCreateImage(vncPtr->display, vi, Tk_Depth(tkwin),
					 ZPixmap, NULL, &vncPtr->shmInfo,
					 SHM_WIDTH, SHM_HEIGHT);
	XSync(vncPtr->display, False);
	if (goterr) {
	    goto noshm_2;
	}
	vncPtr->shmInfo.shmid = shmget(IPC_PRIVATE, SHM_SIZE,
				       IPC_CREAT | 0777);
	vncPtr->shmInfo.shmaddr = (char *) shmat(vncPtr->shmInfo.shmid, 0, 0);
	if (vncPtr->shmInfo.shmaddr == (char *) -1) {
	    shmctl(vncPtr->shmInfo.shmid, IPC_RMID, NULL);
	    goto noshm_1;
	}
	vncPtr->shmInfo.readOnly = False;
	XShmAttach(vncPtr->display, &vncPtr->shmInfo);
	XSync(vncPtr->display, False);
	shmctl(vncPtr->shmInfo.shmid, IPC_RMID, NULL);
	if (goterr) {
	    shmdt(vncPtr->shmInfo.shmaddr);
noshm_1:
	    XDestroyImage(vncPtr->img[1]);
	    vncPtr->img[1] = NULL;
noshm_2:	    
	    Tk_DeleteErrorHandler(handler);
	    goto noshm;
	}
	Tk_DeleteErrorHandler(handler);
	vncPtr->img[1]->data = vncPtr->shmInfo.shmaddr;
    } else {
noshm:
	vncPtr->islocal = 0;
    }
    vncPtr->img[0] = XCreateImage(vncPtr->display, vi, Tk_Depth(tkwin),
				  ZPixmap, 0, NULL, 1, 1, 8, 0);

    RFBInitClient(&vncPtr->rfb, (ClientData) vncPtr, bypp, bps, spp,
		  Tk_Visual(tkwin));
#ifdef HAVE_PTH
    pth_uctx_create(&vncPtr->uc0);
    pth_uctx_create(&vncPtr->uc1);
    vncPtr->uc_run = 0;
    pth_uctx_make(vncPtr->uc1, NULL, 65536, NULL,
		  DoRFBServerMessage, vncPtr, NULL);
#endif

    Tk_CreateEventHandler(new,
			  ExposureMask | StructureNotifyMask |
			  VisibilityChangeMask | PointerMotionMask |
			  ButtonPressMask | ButtonReleaseMask |
			  FocusChangeMask,
			  VNCEventProc, (ClientData) vncPtr);

    if (ConfigureVNC(interp, vncPtr, argc - 2, argv + 2, 0) != TCL_OK) {
	goto error;
    }
    interp->result = Tk_PathName(new);
    if (vncPtr->islocal) {
	Tk_CreateGenericHandler(HandleCompletionEvents, (ClientData) vncPtr);
    }
    vncPtr->flags |= UPDATE_INFO;
    return TCL_OK;

error:
    if (new != NULL) {
	Tk_DestroyWindow(new);
    }
    return TCL_ERROR;
}

/*
 *--------------------------------------------------------------
 *
 * MakeInfo --
 *
 *	Make "array get" information list about
 *	connection state.
 *
 *--------------------------------------------------------------
 */

static void
MakeInfo(vncPtr, dsPtr)
    VNC *vncPtr;
    Tcl_DString *dsPtr;
{
    int i, width = 0, height = 0;
    char buffer[24];

    Tcl_DStringAppendElement(dsPtr, "host");
    if (vncPtr->rfb.serverHost) {
	Tcl_DStringStartSublist(dsPtr);
	Tcl_DStringAppend(dsPtr, vncPtr->rfb.serverHost, -1);
	if (vncPtr->rfb.serverPort != 5900) {
	    sprintf(buffer, ":%d", vncPtr->rfb.serverPort);
	    Tcl_DStringAppend(dsPtr, buffer, -1);
	}
	Tcl_DStringEndSublist(dsPtr);
    } else {
	Tcl_DStringAppendElement(dsPtr, "");
    }
    Tcl_DStringAppendElement(dsPtr, "connected");
    Tcl_DStringAppendElement(dsPtr, vncPtr->rfb.sock == -1 ? "0" : "1");
    Tcl_DStringAppendElement(dsPtr, "desktop");
    Tcl_DStringAppendElement(dsPtr, vncPtr->rfb.sock == -1 ?
			     "" : vncPtr->rfb.desktopName);
    if (vncPtr->rfb.sock != -1) {
	i = vncPtr->canlocal;
	width = vncPtr->img[i]->width;
	height = vncPtr->img[i]->height;
    }
    Tcl_DStringAppendElement(dsPtr, "width");
    sprintf(buffer, "%d", width);
    Tcl_DStringAppendElement(dsPtr, buffer);
    Tcl_DStringAppendElement(dsPtr, "height");
    sprintf(buffer, "%d", height);
    Tcl_DStringAppendElement(dsPtr, buffer);
    Tcl_DStringAppendElement(dsPtr, "viewonly");
    Tcl_DStringAppendElement(dsPtr, vncPtr->rfb.appData.viewOnly ? "1" : "0");
    Tcl_DStringAppendElement(dsPtr, "shared");
    Tcl_DStringAppendElement(dsPtr, vncPtr->rfb.appData.shareDesktop ?
			     "1" : "0");
}

/*
 *--------------------------------------------------------------
 *
 * VNCWidgetCmd --
 *
 *	This procedure is invoked to process the Tcl command
 *	that corresponds to a vnc widget.  See the user
 *	documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

static int
VNCWidgetCmd(clientData, interp, argc, argv)
    ClientData clientData;	/* Information about vnc widget. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int argc;			/* Number of arguments. */
    char **argv;		/* Argument strings. */
{
    VNC *vncPtr = (VNC *) clientData;
    int result = TCL_OK;
#ifdef HAVE_PTH
    int sock = -1;
#endif
    size_t length;
    int i, c;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
			 argv[0], " option ?arg arg ...?\"", (char *) NULL);
	return TCL_ERROR;
    }
    Tcl_Preserve((ClientData) vncPtr);
    c = argv[1][0];
    length = strlen(argv[1]);
    if ((c == 'c') && (strncmp(argv[1], "cget", length) == 0) &&
	(length >= 2)) {
	if (argc != 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " cget option\"", (char *) NULL);
	    result = TCL_ERROR;
	    goto done;
	}
	result = Tk_ConfigureValue(interp, vncPtr->tkwin, configSpecs,
				   (char *) vncPtr, argv[2], 0);
    } else if ((c == 'c') && (strncmp(argv[1], "configure", length) == 0) &&
	       (length >= 4)) {
	if (argc == 2) {
	    result = Tk_ConfigureInfo(interp, vncPtr->tkwin, configSpecs,
				      (char *) vncPtr, (char *) NULL, 0);
	} else if (argc == 3) {
	    result = Tk_ConfigureInfo(interp, vncPtr->tkwin, configSpecs,
				      (char *) vncPtr, argv[2], 0);
	} else {
	    result = ConfigureVNC(interp, vncPtr, argc - 2, argv + 2,
				  TK_CONFIG_ARGV_ONLY);

	}
    } else if ((c == 'c') && (strncmp(argv[1], "connect", length) == 0) &&
	       (length >= 4)) {
	char *p;
	int shared = 0, viewOnly = 0;

	if (argc < 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " connect vnchost ?vncpass ",
			     "?-shared? ?-viewonly??\"",
			     (char *) NULL);
	    result = TCL_ERROR;
	    goto done;
	}
#ifdef HAVE_PTH
	sock = vncPtr->rfb.sock;
	if (vncPtr->uc_run) {
	    vncPtr->uc_run = 0;
	    pth_uctx_switch(vncPtr->uc0, vncPtr->uc1);
	}
	if (sock != -1) {
	    Tcl_DeleteFileHandler(sock);
	    vncPtr->flags |= UPDATE_XVIEW | UPDATE_YVIEW | UPDATE_INFO;
	}
#else
	if (vncPtr->rfb.sock != -1) {
	    Tcl_DeleteFileHandler(vncPtr->rfb.sock);
	    RFBCloseConnection(&vncPtr->rfb);
	    vncPtr->flags |= UPDATE_XVIEW | UPDATE_YVIEW | UPDATE_INFO;
	}
#endif
	if (vncPtr->pwd) {
	    Tcl_Free(vncPtr->pwd);
	    vncPtr->pwd = NULL;
	}
	if (vncPtr->rfb.serverHost) {
	    Tcl_Free(vncPtr->rfb.serverHost);
	}
	vncPtr->rfb.serverHost = Tcl_Alloc(strlen(argv[2]) + 1);
	strcpy(vncPtr->rfb.serverHost, argv[2]);
	p = strchr(vncPtr->rfb.serverHost, ':');
	if (p != NULL) {
	    *p = '\0';
	    ++p;
	    if (Tcl_GetInt(interp, p, &vncPtr->rfb.serverPort)
		!= TCL_OK) {
		result = TCL_ERROR;
		goto done;
	    }
	    if (vncPtr->rfb.serverPort < 5900) {
		vncPtr->rfb.serverPort += 5900;
	    }
	}
	if (argc > 3 && argv[3][0]) {
	    vncPtr->pwd = Tcl_Alloc(strlen(argv[3]) + 1);
	    strcpy(vncPtr->pwd, argv[3]);
	}
	c = 4;
	while (c < argc) {
	    if (!strcmp(argv[c], "-shared")) {
		++shared;
	    } else if (!strcmp(argv[c], "-viewonly")) {
		++viewOnly;
	    } else {
		Tcl_AppendResult(interp, "unknown connect option \"",
			       argv[c], "\"", (char *) NULL);
		result = TCL_ERROR;
		goto done;
	    }
	    ++c;
	}
	vncPtr->needPwd = 0;
#ifdef HAVE_PTH
	vncPtr->rfb.Suspend = NULL;
#endif
	vncPtr->flags |= REDRAW_NOCLIP | UPDATE_XVIEW | UPDATE_YVIEW;
	if (!RFBInitConnection(&vncPtr->rfb, shared, viewOnly)) {
#ifdef HAVE_PTH
	    vncPtr->uc_run = 0;
#endif
	    RFBCloseConnection(&vncPtr->rfb);
	    if (vncPtr->needPwd) {
		Tcl_AppendResult(interp, "password required", (char *) NULL);
	    } else {
		Tcl_AppendResult(interp, "connect failed", (char *) NULL);
	    }
	    result = TCL_ERROR;
	    goto done;
	}
#ifdef HAVE_PTH
	vncPtr->uc_run = 1;
	vncPtr->rfb.Suspend = RFBSwitch;
#endif
	XSetClipMask(vncPtr->display, vncPtr->iGC, None);
	XDestroyRegion(vncPtr->region);
	vncPtr->region = XCreateRegion();
	ComputeView(vncPtr);
	Tcl_CreateFileHandler(vncPtr->rfb.sock, TCL_READABLE,
			      VNCFileHandler, (ClientData) vncPtr);
    } else if ((c == 'd') && (strncmp(argv[1], "disconnect", length) == 0) &&
	       (length >= 1)) {
#ifdef HAVE_PTH
	sock = vncPtr->rfb.sock;
	if (vncPtr->uc_run) {
	    vncPtr->uc_run = 0;
	    pth_uctx_switch(vncPtr->uc0, vncPtr->uc1);
	}
	if (sock != -1) {
	    Tcl_DeleteFileHandler(sock);
	    vncPtr->flags |= UPDATE_XVIEW | UPDATE_YVIEW | UPDATE_INFO;
	}
#else
	if (vncPtr->rfb.sock != -1) {
	    Tcl_DeleteFileHandler(vncPtr->rfb.sock);
	    RFBCloseConnection(&vncPtr->rfb);
	    vncPtr->flags |= UPDATE_XVIEW | UPDATE_YVIEW | UPDATE_INFO;
	}
#endif
	if (!(vncPtr->flags & REDRAW_PENDING)) {
	    Tcl_DoWhenIdle(DisplayVNC, (ClientData) vncPtr);
	    vncPtr->flags |= REDRAW_PENDING;
	}
    } else if ((c == 'i') && (strncmp(argv[1], "info", length) == 0) &&
	       (length >= 1)) {
	Tcl_DString ds;

	Tcl_DStringInit(&ds);
	MakeInfo(vncPtr, &ds);
	Tcl_AppendResult(interp, Tcl_DStringValue(&ds), (char *) NULL);
	Tcl_DStringFree(&ds);
    } else if ((c == 's') && (strncmp(argv[1], "sendkey", length) == 0) &&
	       (length >= 1)) {
	int down, sym;

	if (argc != 4) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " sendkey down sym\"",
			     (char *) NULL);
	    result = TCL_ERROR;
	    goto done;
	}
	if (Tcl_GetBoolean(interp, argv[2], &down) != TCL_OK) {
	    result = TCL_ERROR;
	    goto done;
	}
	if (Tcl_GetInt(interp, argv[3], &sym) != TCL_OK) {
	    result = TCL_ERROR;
	    goto done;
	}
	if ((vncPtr->flags & GOT_FOCUS) && vncPtr->rfb.sock != -1 &&
	    !vncPtr->rfb.appData.viewOnly) {
	    RFBSendKeyEvent(&vncPtr->rfb, sym, down);
	    strcpy(interp->result, "1");
	} else {
	    strcpy(interp->result, "0");
	}
    } else if ((c == 'x') && (strncmp(argv[1], "xview", length) == 0) &&
	       (length >= 1)) {
	i = vncPtr->canlocal;
	if (argc == 2) {
	    char buffer[128];
	    double amt[2];

	    if (vncPtr->viw <= 0 || vncPtr->viw >= vncPtr->img[i]->width) {
		amt[0] = 0;
		amt[1] = 1;
	    } else {
		amt[0] = amt[1] = (double) vncPtr->xoff /
		    (double) vncPtr->img[i]->width;
		amt[1] += (double) vncPtr->viw /
		    (double) vncPtr->img[i]->width;
	    }
	    Tcl_PrintDouble(interp, amt[0], buffer);
	    Tcl_AppendElement(interp, buffer);
	    Tcl_PrintDouble(interp, amt[1], buffer);
	    Tcl_AppendElement(interp, buffer);
	} else if (argc == 3) {
	    int xoff;

	    if (Tcl_GetInt(interp, argv[2], &xoff) == TCL_OK) {
		if (vncPtr->viw < vncPtr->img[i]->width) {
		    if (xoff < 0) {
			xoff = 0;
		    } else if (xoff + vncPtr->viw > vncPtr->img[i]->width) {
			xoff = vncPtr->img[i]->width - vncPtr->viw;
		    }
		    vncPtr->xoff = xoff;
		    ComputeView(vncPtr);
		    vncPtr->flags |= REDRAW_NOCLIP | UPDATE_XVIEW;
		    goto redraw;
		}
	    }
	} else if (argc == 4) {
	    double frac;
	    int xoff;

	    if (strcmp(argv[2], "moveto")) {
		goto badxview;
	    }
	    if (Tcl_GetDouble(interp, argv[3], &frac) == TCL_OK) {
		if (vncPtr->viw < vncPtr->img[i]->width) {
		    xoff = frac * vncPtr->img[i]->width;
		    if (xoff < 0) {
			xoff = 0;
		    } else if (xoff + vncPtr->viw > vncPtr->img[i]->width) {
			xoff = vncPtr->img[i]->width -
			    vncPtr->viw;
		    }
		    vncPtr->xoff = xoff;
		    ComputeView(vncPtr);
		    vncPtr->flags |= REDRAW_NOCLIP | UPDATE_XVIEW;
		    goto redraw;
		}
	    }
	} else if (argc == 5) {
	    int n, isunit = 0;

	    c = argv[4][0];
	    length = strlen(argv[4]);
	    if (strcmp(argv[2], "scroll")) {
		goto badxview;
	    }
	    if ((c == 'u') && (strncmp(argv[4], "units", length) == 0)) {
		isunit = 1;
	    } else if ((c != 'p') || strncmp(argv[4], "pages", length)) {
		goto badxview;
	    }
	    if (Tcl_GetInt(interp, argv[3], &n) == TCL_OK) {
		if (vncPtr->viw < vncPtr->img[i]->width) {
		    int xoff;

		    if (isunit) {
			xoff = vncPtr->xoff + n;
		    } else {
			xoff = vncPtr->xoff + n * vncPtr->img[i]->width / 2;
		    }
		    if (xoff < 0) {
			xoff = 0;
		    } else if (xoff + vncPtr->viw > vncPtr->img[i]->width) {
			xoff = vncPtr->img[i]->width - vncPtr->viw;
		    }
		    vncPtr->xoff = xoff;
		    ComputeView(vncPtr);
		    vncPtr->flags |= REDRAW_NOCLIP | UPDATE_XVIEW;
		    goto redraw;
		}
	    }
	} else {
badxview:
	    Tcl_AppendResult(interp, "bad args to xview subcommand",
			     (char *) NULL);
	    result = TCL_ERROR;
	}
    } else if ((c == 'y') && (strncmp(argv[1], "yview", length) == 0) &&
	       (length >= 1)) {
	i = vncPtr->canlocal;
	if (argc == 2) {
	    char buffer[128];
	    double amt[2];

	    if (vncPtr->vih <= 0 || vncPtr->vih >= vncPtr->img[i]->height) {
		amt[0] = 0;
		amt[1] = 1;
	    } else {
		amt[0] = amt[1] = (double) vncPtr->yoff / 
		    (double) vncPtr->img[i]->height;
		amt[1] += (double) vncPtr->vih /
		    (double) vncPtr->img[i]->height;
	    }
	    Tcl_PrintDouble(interp, amt[0], buffer);
	    Tcl_AppendElement(interp, buffer);
	    Tcl_PrintDouble(interp, amt[1], buffer);
	    Tcl_AppendElement(interp, buffer);
	} else if (argc == 3) {
	    int yoff;

	    if (Tcl_GetInt(interp, argv[2], &yoff) == TCL_OK) {
		if (vncPtr->vih < vncPtr->img[i]->height) {
		    if (yoff < 0) {
			yoff = 0;
		    } else if (yoff + vncPtr->vih > vncPtr->img[i]->height) {
			yoff = vncPtr->img[i]->height - vncPtr->vih;
		    }
		    vncPtr->yoff = yoff;
		    ComputeView(vncPtr);
		    vncPtr->flags |= REDRAW_NOCLIP | UPDATE_YVIEW;
		    goto redraw;
		}
	    }
	} else if (argc == 4) {
	    double frac;
	    int yoff;

	    if (strcmp(argv[2], "moveto")) {
		goto badyview;
	    }
	    if (Tcl_GetDouble(interp, argv[3], &frac) == TCL_OK) {
		if (vncPtr->vih < vncPtr->img[i]->height) {
		    yoff = frac * vncPtr->img[i]->height;
		    if (yoff < 0) {
			yoff = 0;
		    } else if (yoff + vncPtr->vih > vncPtr->img[i]->height) {
			yoff = vncPtr->img[i]->height - vncPtr->vih;
		    }
		    vncPtr->yoff = yoff;
		    ComputeView(vncPtr);
		    vncPtr->flags |= REDRAW_NOCLIP | UPDATE_YVIEW;
		    goto redraw;
		}
	    }
	} else if (argc == 5) {
	    int n, isunit = 0;

	    c = argv[4][0];
	    length = strlen(argv[4]);
	    if (strcmp(argv[2], "scroll")) {
		goto badyview;
	    }
	    if ((c == 'u') && (strncmp(argv[4], "units", length) == 0)) {
		isunit = 1;
	    } else if ((c != 'p') || strncmp(argv[4], "pages", length)) {
		goto badyview;
	    }
	    if (Tcl_GetInt(interp, argv[3], &n) == TCL_OK) {
		if (vncPtr->vih < vncPtr->img[i]->height) {
		    int yoff;

		    if (isunit) {
			yoff = vncPtr->yoff + n;
		    } else {
			yoff = vncPtr->yoff + n * vncPtr->img[i]->height / 2;
		    }
		    if (yoff < 0) {
			yoff = 0;
		    } else if (yoff + vncPtr->vih > vncPtr->img[i]->height) {
			yoff = vncPtr->img[i]->height -
			    vncPtr->vih;
		    }
		    vncPtr->yoff = yoff;
		    ComputeView(vncPtr);
		    vncPtr->flags |= REDRAW_NOCLIP | UPDATE_YVIEW;
		    goto redraw;
		}
	    }
	} else {
badyview:
	    Tcl_AppendResult(interp, "bad args to yview subcommand",
			     (char *) NULL);
	    result = TCL_ERROR;
	}


    } else {
	Tcl_AppendResult(interp, "bad option \"", argv[1],
			 "\": must be cget, configure, connect, etc.",
			 (char *) NULL);
	result = TCL_ERROR;
    }
done:
    Tcl_Release((ClientData) vncPtr);
    return result;

redraw:
    if (!(vncPtr->flags & REDRAW_PENDING)) {
	Tcl_DoWhenIdle(DisplayVNC, (ClientData) vncPtr);
	vncPtr->flags |= REDRAW_PENDING;
    }
    Tcl_Release((ClientData) vncPtr);
    return result;

}

/*
 *----------------------------------------------------------------------
 *
 * DestroyVNC --
 *
 *	This procedure is invoked by Tcl_EventuallyFree or Tcl_Release
 *	to clean up the internal structure of a vnc at a safe time
 *	(when no-one is using it anymore).
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Everything associated with the vnc is freed up.
 *
 *----------------------------------------------------------------------
 */

static void
DestroyVNC(memPtr)
    char *memPtr;		/* Info about vnc widget. */
{
    VNC *vncPtr = (VNC *) memPtr;
    int sock = vncPtr->rfb.sock;

#ifdef HAVE_PTH
    if (vncPtr->uc_run) {
	vncPtr->uc_run = 0;
	pth_uctx_switch(vncPtr->uc0, vncPtr->uc1);
    }
    pth_uctx_destroy(vncPtr->uc0);
    pth_uctx_destroy(vncPtr->uc1);
    if (sock) {
	Tcl_DeleteFileHandler(sock);
    }
#endif
    if (sock) {
	RFBCloseConnection(&vncPtr->rfb);
    }
    if (vncPtr->islocal) {
	Tk_DeleteGenericHandler(HandleCompletionEvents, (ClientData) vncPtr);
	XShmDetach(vncPtr->display, &vncPtr->shmInfo);
	shmdt(vncPtr->shmInfo.shmaddr);
    }
    if (vncPtr->img[0]->data) {
	Tcl_Free((char *) vncPtr->img[0]->data);
    }
    if (vncPtr->rfb.serverHost) {
	Tcl_Free(vncPtr->rfb.serverHost);
    }
    vncPtr->img[0]->data = NULL;
    XDestroyImage(vncPtr->img[0]);
    if (vncPtr->islocal) {
	vncPtr->img[1]->data = NULL;
	XDestroyImage(vncPtr->img[1]);
    }
    XDestroyRegion(vncPtr->region);
    if (vncPtr->iGC != None) {
	XFreeGC(vncPtr->display, vncPtr->iGC);
    }

#ifdef HAVE_LIBZ
    if (vncPtr->rfb.raw_buffer != NULL) {
	free(vncPtr->rfb.raw_buffer);
    }
    if (vncPtr->rfb.decompStreamInited) {
	inflateEnd(&vncPtr->rfb.decompStream);
    }
#endif

#ifdef HAVE_LIBJPEG
    {
	int i;

	for ( i = 0; i < 4; i++ ) {
	    if (vncPtr->rfb.zlibStreamActive[i]) {
		inflateEnd(&vncPtr->rfb.zlibStream[i]);
	    }
	}

	if (vncPtr->rfb.jpegSrcManager) {
	    free(vncPtr->rfb.jpegSrcManager);
	}
    }
#endif

    Tk_FreeOptions(configSpecs, (char *) vncPtr, vncPtr->display, 0);
    Tcl_Free((char *) vncPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * ConfigureVNC --
 *
 *	This procedure is called to process an argv/argc list, plus
 *	the Tk option database, in order to configure (or
 *	reconfigure) a vnc widget.
 *
 * Results:
 *	The return value is a standard Tcl result.  If TCL_ERROR is
 *	returned, then interp->result contains an error message.
 *
 * Side effects:
 *	Configuration information, such as text string, colors, font,
 *	etc. get set for vncPtr;  old resources get freed, if there
 *	were any.
 *
 *----------------------------------------------------------------------
 */

static int
ConfigureVNC(interp, vncPtr, argc, argv, flags)
    Tcl_Interp *interp;		/* Used for error reporting. */
    VNC *vncPtr;		/* Information about widget;  may or may
				 * not already have values for some fields. */
    int argc;			/* Number of valid entries in argv. */
    char **argv;		/* Arguments. */
    int flags;			/* Flags to pass to Tk_ConfigureWidget. */
{
    int width, height;
    unsigned long fillColor = 0;
    XColor *fillColorPtr;
    int dogc = 0;

    fillColorPtr = vncPtr->fillColorPtr;
    if (fillColorPtr) {
	fillColor = fillColorPtr->pixel;
    }
    if (Tk_ConfigureWidget(interp, vncPtr->tkwin, configSpecs,
	    argc, argv, (char *) vncPtr, flags) != TCL_OK) {
	return TCL_ERROR;
    }
    if (vncPtr->fillColorPtr != fillColorPtr) {
	dogc = 1;
    } else if (vncPtr->fillColorPtr &&
	       vncPtr->fillColorPtr->pixel != fillColor) {
	dogc = 1;
    }
    if (dogc) {
	XGCValues gcValues;
	int mask;
	GC newGC;

	if (vncPtr->fillColorPtr) {
	    gcValues.foreground = vncPtr->fillColorPtr->pixel;
	} else {
	    gcValues.foreground = BlackPixel(vncPtr->display,
					     Tk_ScreenNumber(vncPtr->tkwin));
	}
	gcValues.clip_x_origin = 0;
	gcValues.clip_y_origin = 0;
	gcValues.graphics_exposures = True;
	mask = GCForeground | GCClipXOrigin | GCClipYOrigin |
	    GCGraphicsExposures;
	newGC = XCreateGC(vncPtr->display, Tk_WindowId(vncPtr->tkwin),
			  mask, &gcValues);
	if (vncPtr->iGC != None) {
	    XFreeGC(vncPtr->display, vncPtr->iGC);
	}
	vncPtr->iGC = newGC;
	vncPtr->flags |= REDRAW_NOCLIP;
    }
    if (vncPtr->width > 0 || vncPtr->height > 0) {
	width = vncPtr->width + 
	    2 * (vncPtr->highlightWidth + vncPtr->borderWidth);
	height = vncPtr->height + 
	    2 * (vncPtr->highlightWidth + vncPtr->borderWidth);
	Tk_GeometryRequest(vncPtr->tkwin, width, height);
	ComputeView(vncPtr);
	vncPtr->flags |= UPDATE_XVIEW | UPDATE_YVIEW;
    }
    if (Tk_IsMapped(vncPtr->tkwin)) {
	if (!(vncPtr->flags & REDRAW_PENDING)) {
	    Tcl_DoWhenIdle(DisplayVNC, (ClientData) vncPtr);
	    vncPtr->flags |= REDRAW_PENDING;
	}
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * ComputeView --
 *
 *
 *----------------------------------------------------------------------
 */

static void
ComputeView(vncPtr)
    VNC *vncPtr;		/* Information about widget. */
{
    int padding = vncPtr->highlightWidth + vncPtr->borderWidth;
    int i, width, height;

    width = Tk_Width(vncPtr->tkwin) -
	2 * (vncPtr->highlightWidth + vncPtr->borderWidth);
    height = Tk_Height(vncPtr->tkwin) -
	2 * (vncPtr->highlightWidth + vncPtr->borderWidth);

    i = vncPtr->canlocal;
    if (vncPtr->img[i]->data && vncPtr->rfb.sock != -1 &&
	width > 0 && height > 0) {
	int dx, dy, iw, ih;

	iw = vncPtr->img[i]->width;
	ih = vncPtr->img[i]->height;
	dx = dy = padding;
	if (width > iw) {
	    dx = (width - iw) / 2;
	    vncPtr->xoff = 0;
	}
	if (height > ih) {
	    dy = (height - ih) / 2;
	    vncPtr->yoff = 0;
	}
	if (iw > width) {
	    iw = width;
	}
	if (ih > height) {
	    ih = height;
	}
	vncPtr->vxoff = dx;
	vncPtr->vyoff = dy;
	vncPtr->viw = iw;
	vncPtr->vih = ih;
	if (vncPtr->viw + vncPtr->xoff > vncPtr->img[i]->width) {
	    vncPtr->xoff = vncPtr->img[i]->width - vncPtr->viw;
	}
	if (vncPtr->vih + vncPtr->yoff > vncPtr->img[i]->height) {
	    vncPtr->yoff = vncPtr->img[i]->height - vncPtr->vih;
	}
    } else {
	vncPtr->vxoff = vncPtr->vyoff = padding;
	vncPtr->viw = vncPtr->vih = 0;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * DisplayVNC --
 *
 *	This procedure is invoked to display a vnc widget.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Commands are output to X to display the vnc in its
 *	current mode.
 *
 *----------------------------------------------------------------------
 */

static void
DisplayVNC(clientData)
    ClientData clientData;	/* Information about widget. */
{
    VNC *vncPtr = (VNC *) clientData;
    Tk_Window tkwin = vncPtr->tkwin;
    int i, width, height, padding;

    vncPtr->flags &= ~REDRAW_PENDING;
    if (tkwin == NULL || !Tk_IsMapped(tkwin)) {
	vncPtr->flags &= ~(REDRAW_XEVENTS | REDRAW_NOCLIP);
	XSetClipMask(vncPtr->display, vncPtr->iGC, None);
	XDestroyRegion(vncPtr->region);
	vncPtr->region = XCreateRegion();
	return;
    }
    Tcl_Preserve((ClientData) vncPtr);
    if (vncPtr->flags & REDRAW_XEVENTS) {
	if (vncPtr->borderWidth > 0) {
	    Tk_Draw3DRectangle(tkwin, Tk_WindowId(tkwin), vncPtr->border,
			       vncPtr->highlightWidth,
			       vncPtr->highlightWidth,
			       Tk_Width(tkwin) - 2 * vncPtr->highlightWidth,
			       Tk_Height(tkwin) - 2 * vncPtr->highlightWidth,
			       vncPtr->borderWidth, vncPtr->relief);
	}
	if (vncPtr->highlightWidth > 0) {
	    XRectangle rects[4];
	    GC gc;

	    padding = 0;
	    width = vncPtr->highlightWidth;
	    if (vncPtr->flags & GOT_FOCUS) {
		gc = Tk_GCForColor(vncPtr->highlightColorPtr,
				   Tk_WindowId(tkwin));
	    } else {
		gc = Tk_GCForColor(vncPtr->highlightBgColorPtr,
				   Tk_WindowId(tkwin));
	    }
	    rects[0].x = padding;
	    rects[0].y = padding;
	    rects[0].width = Tk_Width(tkwin) - (2 * padding);
	    rects[0].height = width;
	    rects[1].x = padding;
	    rects[1].y = Tk_Height(tkwin) - width - padding;
	    rects[1].width = Tk_Width(tkwin) - (2 * padding);
	    rects[1].height = width;
	    rects[2].x = padding;
	    rects[2].y = width + padding;
	    rects[2].width = width;
	    rects[2].height = Tk_Height(tkwin) - 2*width - 2*padding;
	    rects[3].x = Tk_Width(tkwin) - width - padding;
	    rects[3].y = rects[2].y;
	    rects[3].width = width;
	    rects[3].height = rects[2].height;
	    XFillRectangles(vncPtr->display, Tk_WindowId(tkwin), gc, rects, 4);
	}
    }

    padding = vncPtr->highlightWidth + vncPtr->borderWidth;
    width = Tk_Width(tkwin) - 2 * padding;
    height = Tk_Height(tkwin) - 2 * padding;

    i = vncPtr->canlocal;
    if (vncPtr->img[i]->data && vncPtr->rfb.sock != -1) {
	if (i) {
	    if (vncPtr->flags & COMPL_PENDING) {
		vncPtr->flags |= REDRAW_AGAIN;
	    } else {
		if (vncPtr->flags & REDRAW_NOCLIP) {
		    XSetClipMask(vncPtr->display, vncPtr->iGC, None);
		} else {
		    XSetRegion(vncPtr->display, vncPtr->iGC, vncPtr->region);
		}
		XDestroyRegion(vncPtr->region);
		vncPtr->region = XCreateRegion();
		XShmPutImage(vncPtr->display,
			     Tk_WindowId(tkwin), vncPtr->iGC,
			     vncPtr->img[1], vncPtr->xoff, vncPtr->yoff,
			     vncPtr->vxoff, vncPtr->vyoff,
			     vncPtr->viw, vncPtr->vih, True);
		vncPtr->flags |= COMPL_PENDING;
	    }
	} else {
	    if (vncPtr->flags & REDRAW_NOCLIP) {
		XSetClipMask(vncPtr->display, vncPtr->iGC, None);
	    } else {
		XSetRegion(vncPtr->display, vncPtr->iGC, vncPtr->region);
	    }
	    XDestroyRegion(vncPtr->region);
	    vncPtr->region = XCreateRegion();
	    XPutImage(vncPtr->display, Tk_WindowId(tkwin), vncPtr->iGC,
		      vncPtr->img[0], vncPtr->xoff, vncPtr->yoff,
		      vncPtr->vxoff, vncPtr->vyoff,
		      vncPtr->viw, vncPtr->vih);
	}
	if (vncPtr->vxoff > padding) {
	    XFillRectangle(vncPtr->display, Tk_WindowId(tkwin),
			   vncPtr->iGC, padding, padding,
			   vncPtr->vxoff - padding, height);
	}
	if (vncPtr->vxoff + vncPtr->viw < width + padding) {
	    int tmp = vncPtr->vxoff + vncPtr->viw;

	    XFillRectangle(vncPtr->display, Tk_WindowId(tkwin),
			   vncPtr->iGC, tmp, padding,
			   width + padding - tmp, height);
	}
	if (vncPtr->vyoff > padding) {
	    XFillRectangle(vncPtr->display, Tk_WindowId(tkwin),
			   vncPtr->iGC, padding, padding,
			   width, vncPtr->vyoff - padding);
	}
	if (vncPtr->vyoff + vncPtr->vih < height + padding) {
	    int tmp = vncPtr->vyoff + vncPtr->vih;

	    XFillRectangle(vncPtr->display, Tk_WindowId(tkwin),
			   vncPtr->iGC, padding, tmp,
			   width, height + padding - tmp);
	}
    } else {
	XSetClipMask(vncPtr->display, vncPtr->iGC, None);
	XDestroyRegion(vncPtr->region);
	vncPtr->region = XCreateRegion();
	if (width > 0 && height > 0) {
	    XFillRectangle(vncPtr->display, Tk_WindowId(tkwin), vncPtr->iGC,
			   vncPtr->highlightWidth + vncPtr->borderWidth,
			   vncPtr->highlightWidth + vncPtr->borderWidth,
			   width, height);
	}
    }
    if ((vncPtr->flags & UPDATE_XVIEW) && vncPtr->xScrollCmd) {
	Tcl_Interp *interp = vncPtr->interp;
	char buffer[128];
	double amt[2];
	int result;

	if (vncPtr->viw <= 0 || vncPtr->viw >= vncPtr->img[i]->width) {
	    amt[0] = 0;
	    amt[1] = 1;
	} else {
	    amt[0] = amt[1] = (double) vncPtr->xoff /
		(double) vncPtr->img[i]->width;
	    amt[1] += (double) vncPtr->viw /
		(double) vncPtr->img[i]->width;
	}
	sprintf(buffer, " %g %g", amt[0], amt[1]);
	Tcl_Preserve(interp);
	result = Tcl_VarEval(interp, vncPtr->xScrollCmd, buffer,
			     (char *) NULL);
	if (result != TCL_OK) {
	    Tcl_AddErrorInfo(interp,
			     "\n    (xscrollcommand executed by vnc)");
	    Tcl_BackgroundError(interp);
	}
	Tcl_Release((ClientData) interp);
    }
    if ((vncPtr->flags & UPDATE_YVIEW) && vncPtr->yScrollCmd) {
	Tcl_Interp *interp = vncPtr->interp;
	char buffer[128];
	double amt[2];
	int result;

	if (vncPtr->vih <= 0 || vncPtr->vih >= vncPtr->img[i]->height) {
	    amt[0] = 0;
	    amt[1] = 1;
	} else {
	    amt[0] = amt[1] = (double) vncPtr->yoff /
		(double) vncPtr->img[i]->height;
	    amt[1] += (double) vncPtr->vih /
		(double) vncPtr->img[i]->height;
	}
	sprintf(buffer, " %g %g", amt[0], amt[1]);
	Tcl_Preserve(interp);
	result = Tcl_VarEval(interp, vncPtr->yScrollCmd, buffer,
			     (char *) NULL);
	if (result != TCL_OK) {
	    Tcl_AddErrorInfo(interp,
			     "\n    (yscrollcommand executed by vnc)");
	    Tcl_BackgroundError(interp);
	}
	Tcl_Release((ClientData) interp);
    }
    if ((vncPtr->flags & UPDATE_INFO) && vncPtr->infoCmd) {
	Tcl_Interp *interp = vncPtr->interp;
	Tcl_DString ds;
	int result;

	Tcl_DStringInit(&ds);
	Tcl_DStringStartSublist(&ds);
	MakeInfo(vncPtr, &ds);
	Tcl_DStringEndSublist(&ds);
	Tcl_Preserve(interp);
	result = Tcl_VarEval(interp, vncPtr->infoCmd, " ",
			     Tcl_DStringValue(&ds), (char *) NULL);
	if (result != TCL_OK) {
	    Tcl_AddErrorInfo(interp,
			     "\n    (infocommand executed by vnc)");
	    Tcl_BackgroundError(interp);
	}
	Tcl_Release((ClientData) interp);
	Tcl_DStringFree(&ds);
    }
    vncPtr->flags &= ~(REDRAW_XEVENTS | REDRAW_NOCLIP |
		       UPDATE_XVIEW | UPDATE_YVIEW | UPDATE_INFO);
    Tcl_Release((ClientData) vncPtr);
}

/*
 *--------------------------------------------------------------
 *
 * VNCEventProc --
 *
 *	This procedure is invoked by the Tk dispatcher on
 *	structure changes to a vnc.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	When the window gets deleted, internal structures get
 *	cleaned up.  When it gets exposed, it is redisplayed.
 *
 *--------------------------------------------------------------
 */

static void
VNCEventProc(clientData, eventPtr)
    ClientData clientData;	/* Information about window. */
    XEvent *eventPtr;		/* Information about event. */
{
    VNC *vncPtr = (VNC *) clientData;
    int i, redraw = 0;

    if (vncPtr->tkwin == NULL) {
	return;
    }
    i = vncPtr->canlocal;
    switch (eventPtr->type) {
    case Expose: {
	XRectangle rect;

	rect.x = eventPtr->xexpose.x;
	rect.y = eventPtr->xexpose.y;
	rect.width = eventPtr->xexpose.width;
	rect.height = eventPtr->xexpose.height;
	XUnionRectWithRegion(&rect, vncPtr->region, vncPtr->region);
	if (eventPtr->xexpose.count == 0) {
	    redraw = 1;
	}
	break;
    }
    case MapNotify:
    case ConfigureNotify:
	vncPtr->flags |= REDRAW_NOCLIP | UPDATE_XVIEW | UPDATE_YVIEW;
	ComputeView(vncPtr);
	redraw = 1;
	break;
    case DestroyNotify:
	if (vncPtr->tkwin != NULL) {
	    Tk_DeleteEventHandler(vncPtr->tkwin,
				  ExposureMask | StructureNotifyMask |
				  VisibilityChangeMask | PointerMotionMask |
				  ButtonPressMask | ButtonReleaseMask |
				  FocusChangeMask,
				  VNCEventProc, (ClientData) vncPtr);
	    vncPtr->tkwin = NULL;
	    Tcl_DeleteCommandFromToken(vncPtr->interp, vncPtr->widgetCmd);
	    vncPtr->widgetCmd = NULL;
	}
	if (vncPtr->flags & REDRAW_PENDING) {
	    Tcl_CancelIdleCall(DisplayVNC, (ClientData) vncPtr);
	}
	Tcl_EventuallyFree((ClientData) vncPtr, DestroyVNC);
	break;
    case FocusIn:
	if (eventPtr->xfocus.detail != NotifyInferior) {
	    if (!(vncPtr->flags & GOT_FOCUS)) {
		vncPtr->flags |= GOT_FOCUS;
		redraw = vncPtr->highlightWidth > 0;
	    }
	}
	break;
    case FocusOut:
	if (eventPtr->xfocus.detail != NotifyInferior) {
	    if (vncPtr->flags & GOT_FOCUS) {
		vncPtr->flags &= ~GOT_FOCUS;
		redraw = vncPtr->highlightWidth > 0;
	    }
	}
	break;
    case ButtonPress:
    case ButtonRelease:
	if ((vncPtr->flags & GOT_FOCUS) && vncPtr->rfb.sock != -1 &&
	    !vncPtr->rfb.appData.viewOnly) {
	    int mask = (eventPtr->xmotion.state >> 8) & 0x1f;
	    int x, y;

	    if (eventPtr->type == ButtonPress) {
		mask |= 1 << (eventPtr->xbutton.button - 1); 
	    } else {
		mask &= ~(1 << (eventPtr->xbutton.button - 1)); 
	    }
	    x = eventPtr->xbutton.x - vncPtr->vxoff + vncPtr->xoff;
	    y = eventPtr->xbutton.y - vncPtr->vyoff + vncPtr->yoff;
	    if (x < 0) {
		x = 0;
	    } else if (x >= vncPtr->img[i]->width) {
		x = vncPtr->img[i]->width - 1;
	    }
	    if (y < 0) {
		y = 0;
	    } else if (y >= vncPtr->img[i]->height) {
		y = vncPtr->img[i]->height - 1;
	    }
	    RFBSendPointerEvent(&vncPtr->rfb, x, y, mask);
	}
	break;
    case MotionNotify:
	if ((vncPtr->flags & GOT_FOCUS) && vncPtr->rfb.sock != -1 &&
	    !vncPtr->rfb.appData.viewOnly) {
	    int mask = (eventPtr->xmotion.state >> 8) & 0x1f;
	    int x, y;
	    x = eventPtr->xbutton.x - vncPtr->vxoff + vncPtr->xoff;
	    y = eventPtr->xbutton.y - vncPtr->vyoff + vncPtr->yoff;
	    if (x < 0) {
		x = 0;
	    } else if (x >= vncPtr->img[i]->width) {
		x = vncPtr->img[i]->width - 1;
	    }
	    if (y < 0) {
		y = 0;
	    } else if (y >= vncPtr->img[i]->height) {
		y = vncPtr->img[i]->height - 1;
	    }
	    RFBSendPointerEvent(&vncPtr->rfb, x, y, mask);
	}
	break;
    }
    if (redraw) {
	vncPtr->flags |= REDRAW_XEVENTS;
	if (!(vncPtr->flags & REDRAW_PENDING)) {
	    Tcl_DoWhenIdle(DisplayVNC, (ClientData) vncPtr);
	    vncPtr->flags |= REDRAW_PENDING;
	}
    }
}

/*
 *----------------------------------------------------------------------
 *
 * VNCCmdDeletedProc --
 *
 *	This procedure is invoked when a widget command is deleted.  If
 *	the widget isn't already in the process of being destroyed,
 *	this command destroys it.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The widget is destroyed.
 *
 *----------------------------------------------------------------------
 */

static void
VNCCmdDeletedProc(clientData)
    ClientData clientData;	/* Pointer to widget record for widget. */
{
    VNC *vncPtr = (VNC *) clientData;
    Tk_Window tkwin = vncPtr->tkwin;

    /*
     * This procedure could be invoked either because the window was
     * destroyed and the command was then deleted (in which case tkwin
     * is NULL) or because the command was deleted, and then this procedure
     * destroys the widget.
     */

    if (tkwin != NULL) {
	vncPtr->tkwin = NULL;
	Tk_DestroyWindow(tkwin);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Vnc_Init --
 *
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The "vnc" Tcl command is created and the "vnc" package
 *	is registered.
 *
 *----------------------------------------------------------------------
 */

int
Vnc_Init(interp)
    Tcl_Interp *interp;		/* Interpreter in which to create command. */
{
    Tk_Window mainwin;

#ifdef USE_TCL_STUBS
    if (!Tcl_InitStubs(interp, "8.0", 0)) {
	return TCL_ERROR;
    }
#else
    if (Tcl_PkgRequire(interp, "Tk", TK_VERSION, 0) == NULL) {
	return TCL_ERROR;
    }
#endif
#ifdef USE_TK_STUBS
    if (!Tk_InitStubs(interp, "8.0", 0)) {
	return TCL_ERROR;
    }
#else
    if (Tcl_PkgRequire(interp, "Tk", TK_VERSION, 0) == NULL) {
	return TCL_ERROR;
    }
#endif
    mainwin = Tk_MainWindow(interp);
    if (mainwin == NULL) {
	return TCL_ERROR;
    }
    if (Tcl_PkgProvide(interp, "vnc", "1.0") != TCL_OK) {
	return TCL_ERROR;
    }
    if (Tcl_CreateCommand(interp, "vnc", VNCCmd, (ClientData) mainwin,
			  (Tcl_CmdDeleteProc *) NULL) == NULL) {
	return TCL_ERROR;
    }
    Tcl_GlobalEval(interp,
		   "bind VNC <Control-Tab> continue\n"
		   "bind VNC <Any-KeyPress> {%W sendkey 1 %N ; break}\n"
		   "bind VNC <Any-KeyRelease> {%W sendkey 0 %N ; break}\n"
		   );
    return TCL_OK;
}
